
#include <avr/io.h>
#include <util/delay.h>
#include <stdlib.h>

#include "u8glib/u8g.h"

#define DELAY_1	1

#define WIDTH  128
#define HEIGHT 64

#define UP   			0
#define DOWN  			1
#define LEFT  			2
#define RIGHT 			3
#define STOP  			5

#define BLOCK_SIZE 		2

u8g_t u8g;				// LCD

int8_t x, y;
int8_t state;          // 0 or 1

int steps, colisions;



/* Set ports directions */
void init_buttons(void) 
{
    DDRD = 0x00;				// Set port D as input port
    PORTD = 0xff;				// Activate pull-up resistors on port D   
	
	DDRD |= (1 << PD7);
	PORTD &= ~(1 << PD7);
    
}

/* Initializam lcd */
void init_lcd(void) 
{
    MCUCR |= (1 << JTD); 
    MCUCR |= (1 << JTD);
    
    // init LCD - data ports PB0 - PB7, control on PA0 - PA4
    u8g_Init8Bit(&u8g, &u8g_dev_st7920_128x64_4x_8bit,
            PN(1, 0), PN(1, 1), PN(1, 2), PN(1, 3), 
            PN(1, 4), PN(1, 5), PN(1, 6), PN(1, 7), 		// DATA PORTS
            PN(0, 2),    									// ENABLE
            U8G_PIN_NONE,
            U8G_PIN_NONE,
            PN(0, 0),    									// RS / DI
            PN(0, 1),    									// RW
            PN(0, 3));   									// RESET
}


void draw_initial_window(void) 
{
    u8g_SetDefaultForegroundColor(&u8g);
    u8g_SetFont(&u8g, u8g_font_helvB10);
    u8g_DrawStr(&u8g, 5, 25, "Level 1 - Joc Nou!");
	u8g_DrawStr(&u8g, 5, 40, "Apasati un Buton.");
}

void draw_won_window(void) 
{
    u8g_SetDefaultForegroundColor(&u8g);
    u8g_SetFont(&u8g, u8g_font_helvB10);
    u8g_DrawStr(&u8g, 5, 25, "Ati Castigat!");
	u8g_DrawStr(&u8g, 5, 40, "Apasati un Buton.");
}


char str[40];
int magic1;

int Xstart0[] = {0,   0, 67, 61,  0};
int Ystart0[] = {10, 20, 10, 20, 10};
int Xend0[]   = {67, 61, 67, 61,  0};
int Yend0[]   = {10, 20, 43, 43, 20};
int nrSeg0 = 5;

int Xstart1[] = {0,  90, 100, 120, 100,  0,  0, 61, 67, 90, 67, 67, 100, 120, 100, 100, 127,  90, 120,  0,  67};
int Ystart1[] = {10,  0,   0,   0,  10, 20, 43, 20, 20, 20, 20, 30,  20,  20,  20,  30,   0,   0,   0, 10,  43};
int Xend1[]   = {90, 90, 100, 120, 120, 61, 61, 61, 67, 90, 90, 90, 100, 120, 120, 120, 127, 100, 127,  0, 127};
int Yend1[]   = {10, 10,  10,  10,  10, 20, 43, 43, 30, 30, 20, 30,  30,  30,  20,  30,  43,   0,   0, 20,  43};
int nrSeg1 = 21;

void draw_game0(void) 
{
    u8g_SetDefaultForegroundColor(&u8g);
	
	sprintf(str, "st: %d cns: %d\0", steps, colisions);
	u8g_DrawStr(&u8g, 2, HEIGHT - 2, str);
	
	// todo draw obstacles
	int i;
	for (i = 0; i < nrSeg0; i++) {
		u8g_DrawLine(&u8g, Xstart0[i], Ystart0[i], Xend0[i], Yend0[i]);
	}
	
	// limit of the play board
	u8g_DrawLine(&u8g, 0, HEIGHT - 20, WIDTH/2 - 3, HEIGHT - 20);
	u8g_DrawLine(&u8g, WIDTH/2 + 3, HEIGHT - 20, WIDTH, HEIGHT - 20);
	
	u8g_DrawBox(&u8g, x - BLOCK_SIZE / 2, y - BLOCK_SIZE / 2, BLOCK_SIZE, BLOCK_SIZE);
	
}

void draw_game1(void) 
{
    u8g_SetDefaultForegroundColor(&u8g);
	
	sprintf(str, "st: %d cns: %d\0", steps, colisions);
	u8g_DrawStr(&u8g, 2, HEIGHT - 2, str);
	
	// todo draw obstacles
	int i;
	for (i = 0; i < nrSeg1; i++) {
		u8g_DrawLine(&u8g, Xstart1[i], Ystart1[i], Xend1[i], Yend1[i]);
	}
	
	// limit of the play board
	u8g_DrawLine(&u8g, 0, HEIGHT - 20, WIDTH/2 - 3, HEIGHT - 20);
	u8g_DrawLine(&u8g, WIDTH/2 + 3, HEIGHT - 20, WIDTH, HEIGHT - 20);
	
	u8g_DrawBox(&u8g, x - BLOCK_SIZE / 2, y - BLOCK_SIZE / 2, BLOCK_SIZE, BLOCK_SIZE);
	
}

/* verificam intrarile */
uint8_t grab_input(void) 
{

    // UP
    if (!(PIND & (1 << PD0))) {
        return UP;
    }
    
    // LEFT
    if (!(PIND & (1 << PD1))) {
        return LEFT;
    }
    
    // RIGHT
    if (!(PIND & (1 << PD4))) {
        return RIGHT;
    }
    
    // DOWN
    if (!(PIND & (1 << PD5))) {
        return DOWN;
    }
	
	return STOP;
}

#define pow2(x) ((x)*(x))

float dist2D(int x, int y, int X, int Y)
{
	int v1 = pow2(x - X);
	int v2 = pow2(y - Y);
	int v = v1 + v2;
	
	return sqrt(v);
}

#define max2(a, b) ((a > b) ? a : b)
#define min2(a, b) ((a < b) ? a : b)

uint8_t isOnBorder0(int x, int y)
{
	for (int i = 0; i < nrSeg0; i++) {
		
		int xMax = max2(Xstart0[i], Xend0[i]);
		int xMin = min2(Xstart0[i], Xend0[i]);
		
		int yMax = max2(Ystart0[i], Yend0[i]);
		int yMin = min2(Ystart0[i], Yend0[i]);
		
		
		if (xMax != xMin && yMax != yMin) {
			if (!(xMin <= x && xMax >= x)) {
				continue;
			}
			
			if (!(yMin <= y && yMax >= y)) {
				continue;
			}
		} else {
			if (xMax == xMin) {
				if (!(yMin <= y && yMax >= y)) {
					continue;
				}
			} else {
				if (!(xMin <= x && xMax >= x)) {
					continue;
				}
			}
		}
		
		
		
		// https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line
		int val1 = (Yend0[i] - Ystart0[i]) * x - (Xend0[i] - Xstart0[i]) * y;
		val1 = val1 + Xend0[i] * Ystart0[i] - Yend0[i] * Xstart0[i];
		val1 = abs(val1);
		
		int val2 = pow2(Yend0[i] - Ystart0[i]) + pow2(Xend0[i] - Xstart0[i]);
		double tmp = sqrt(val2);
		
		int dist = (int) ((1.0 * val1) / tmp);
		
		if (dist < magic1) {
			magic1 = dist;
		}
		
		if (dist <= 1) {
			return 1;
		}
		
	}
	
	return 0;
}

uint8_t isOnBorder1(int x, int y)
{
	for (int i = 0; i < nrSeg1; i++) {
		
		int xMax = max2(Xstart1[i], Xend1[i]);
		int xMin = min2(Xstart1[i], Xend1[i]);
		
		int yMax = max2(Ystart1[i], Yend1[i]);
		int yMin = min2(Ystart1[i], Yend1[i]);
		
		
		if (xMax != xMin && yMax != yMin) {
			if (!(xMin <= x && xMax >= x)) {
				continue;
			}
			
			if (!(yMin <= y && yMax >= y)) {
				continue;
			}
		} else {
			if (xMax == xMin) {
				if (!(yMin <= y && yMax >= y)) {
					continue;
				}
			} else {
				if (!(xMin <= x && xMax >= x)) {
					continue;
				}
			}
		}
		
		
		
		
		int val1 = (Yend1[i] - Ystart1[i]) * x - (Xend1[i] - Xstart1[i]) * y;
		val1 = val1 + Xend1[i] * Ystart1[i] - Yend1[i] * Xstart1[i];
		val1 = abs(val1);
		
		int val2 = pow2(Yend1[i] - Ystart1[i]) + pow2(Xend1[i] - Xstart1[i]);
		double tmp = sqrt(val2);
		
		int dist = (int) ((1.0 * val1) / tmp);
		
		if (dist < magic1) {
			magic1 = dist;
		}
		
		if (dist <= 1) {
			return 1;
		}
		
	}
	
	return 0;
}

uint8_t gameWon()
{
	if (x > 61 && x < 67 && y >= 43) {
		return 1;
	}
	
	return 0;
}

void makeSound(void)
{
	for (int i = 0; i < 500; i++) {
		PORTD |= (1 << PD7);
		_delay_ms(0.1);
		PORTD &= ~(1 << PD7);
		_delay_ms(0.1);
	}
}

void makeColisionSound(void)
{
	for (int i = 0; i < 4000; i++) {
		PORTD |= (1 << PD7);
		_delay_ms(0.1);
		PORTD &= ~(1 << PD7);
		_delay_ms(0.1);
	}
}

int main(void) 
{
	
begin:	

    init_buttons();
    init_lcd();
    
    // asteptam pana la prima comanda pentru a porni jocul
    while(1) {
        uint8_t dir = grab_input();
		
		if (dir != STOP) break;
        
        u8g_FirstPage(&u8g);
        do {
            draw_initial_window();
        } while (u8g_NextPage(&u8g));    
        
        _delay_ms(DELAY_1);
    }
	
	makeSound();
	
	x = y = 15;
	state = 1;
	
	steps = 0;
	colisions = 0;
	
	while (1) {
        uint8_t dir = grab_input(); 
		
		if (dir != STOP) {
			int xOld = x;
			int yOld = y;
			if (dir == UP) {
				y = y + BLOCK_SIZE;
				if (y < 0) y = 0;
				if (y > HEIGHT - 20) y = HEIGHT - 20;
			} else if (dir == DOWN) {
				y = y - BLOCK_SIZE;
				if (y < 0) y = 0;
				if (y > HEIGHT - 20) y = HEIGHT - 20;
			} else if (dir == LEFT) {
				x = x - BLOCK_SIZE;
				if (x < 0) x = 0;
				if (x > WIDTH - 1) x = WIDTH - 1;
			} else if (dir == RIGHT) {
				x = x + BLOCK_SIZE;
				if (x < 0) x = 0;
				if (x > WIDTH - 1) x = WIDTH - 1;
			}
			
			magic1 = 10;
			uint8_t onBorder = isOnBorder0(x, y);
			
			if (onBorder) {
				x = xOld;
				y = yOld;
				colisions++;
				makeColisionSound();
			} else {
				steps++;
			}
		}
		
			
        u8g_FirstPage(&u8g);
        do {
            draw_game0();
        } while (u8g_NextPage(&u8g)); 
		
		if (gameWon()) {
			goto level1;
		}
		
		_delay_ms(DELAY_1);
    }
	
	
	
level1:	
	makeSound();
	x = y = 15;
	state = 1;
	
	steps = 0;
	colisions = 0;
	
	while (1) {
        uint8_t dir = grab_input(); 
		
		if (dir != STOP) {
			int xOld = x;
			int yOld = y;
			if (dir == UP) {
				y = y + BLOCK_SIZE;
				if (y < 0) y = 0;
				if (y > HEIGHT - 20) y = HEIGHT - 20;
			} else if (dir == DOWN) {
				y = y - BLOCK_SIZE;
				if (y < 0) y = 0;
				if (y > HEIGHT - 20) y = HEIGHT - 20;
			} else if (dir == LEFT) {
				x = x - BLOCK_SIZE;
				if (x < 0) x = 0;
				if (x > WIDTH - 1) x = WIDTH - 1;
			} else if (dir == RIGHT) {
				x = x + BLOCK_SIZE;
				if (x < 0) x = 0;
				if (x > WIDTH - 1) x = WIDTH - 1;
			}
			
			magic1 = 10;
			uint8_t onBorder = isOnBorder1(x, y);
			
			if (onBorder) {
				x = xOld;
				y = yOld;
				colisions++;
				makeColisionSound();
			} else {
				steps++;
			}
		}
		
			
        u8g_FirstPage(&u8g);
        do {
            draw_game1();
        } while (u8g_NextPage(&u8g)); 
		
		if (gameWon()) {
			goto game_won;
		}
		
		_delay_ms(DELAY_1);
    }
 
game_won: 
	makeSound();
	while(1) {
        uint8_t dir = grab_input();
		
		if (dir != STOP) break;
        
        u8g_FirstPage(&u8g);
        do {
            draw_won_window();
        } while (u8g_NextPage(&u8g));    
        
        _delay_ms(DELAY_1);
    }

	goto begin;
	
    return 0;
}